/*******************************************
ALU (Arithmetic and Logic Unit)
SR (Status Register)
UNIT
********************************************
HOMS
Hardware-Oriented Microprocessor
Simulator
(C) Panayotis (Panos) Papazoglou
LICENSE: 
Creative Commons
CC BY NC SA
International License
********************************************/
//UNIT ADDRESS = 9
#define ADDRESS 9

//Include I2C library
#include <Wire.h>

//Include LCD library
#include <LiquidCrystal.h>

//Set pins for LCD1602 connection
LiquidCrystal lcd(2,3,4,5,6,7);

//This UNIT Receives/Sends commands from/to other units as bytes 
#define READ  65
#define ADD 72

//Variable that is used inside ISR
volatile byte rvTRUE=0;

//Byte received from other units
byte rV;

//Store result of the numerical calculation
byte result=0;

//Store the content of temporarily registers T1 and T2
byte T1=0; 
byte T2=0;

//Flag array
byte SR[]={0,0,0}; //VF (oVerflow), ZF (Zero), SF (Sign)

//Starting function (after reset or power on), runs once
void setup() 
{
  //Start I2C communication
  Wire.begin(ADDRESS);

  //Activate Receive & Request ISR routines
  Wire.onReceive(onReceive);
  Wire.onRequest(onRequest);

  //Initialize LCD screen
  lcd.begin(16,2);

  //Display result and flag status
  DISPLAY_RES();
}

void loop() 
{
  //If a byte is received and is an ADD command,
  //then execute addition between the contents
  //of special registers T1 and T2
  //Store result in a variable, update SR[] array
  //and display result/SR contents
  if (rvTRUE==1)
  {
      if (rV==ADD) 
        {
            T1=READ_REG(7);
            delay(100);
            T2=READ_REG(8);
            result=T1+T2;
            UPDATE_SR(result);
            DISPLAY_RES();
        }
      rvTRUE=0;
  }
}

/*******************************************
ISR routine
Is activated when a byte is received
via I2C communication
********************************************/
void onReceive(int a) 
{ 
  rV=Wire.read();

  //Flag for activating code in order to read T1 and T2 contents
  //I2C functions can not be called within an active I2C routine 
  rvTRUE=1;
}

/*******************************************
ISR routine
Is activated when a request is received
via I2C communication
********************************************/
void onRequest()
{
  //Send result (ALU calculation) as answer to an I2C request
  Wire.write(result);
}

/*******************************************
Read the content of a Register
with address addr
********************************************/
byte READ_REG(byte addr)
{
byte ans;

  Wire.beginTransmission(addr);
  Wire.write(READ);
  Wire.endTransmission();

  //Receive answer
  Wire.requestFrom(addr, 1);
  if (Wire.available()) 	      
		ans = Wire.read();

return ans;
}

/*******************************************
Update Status Register Flags based
on result status
********************************************/
void UPDATE_SR(byte res)
{

if (res==0) SR[1]=1; else SR[1]=0;
if (res<0) SR[2]=1; else SR[2]=0;
if (res>255) {SR[0]=1; result=0;} else SR[0]=0;
}

/*******************************************
Display result and Status Register Flags 
on LCD screen
********************************************/
void DISPLAY_RES()
{
  lcd.setCursor(0,0); lcd.print("RES10="); lcd.setCursor(6,0); lcd.print(result);
  lcd.setCursor(10,0); lcd.print("HEX="); lcd.setCursor(14,0); lcd.print(String(result,HEX));
  lcd.setCursor(4,1); lcd.print("V:"); lcd.setCursor(6,1); lcd.print(SR[0]);
  lcd.setCursor(8,1); lcd.print("Z:"); lcd.setCursor(10,1); lcd.print(SR[1]);
  lcd.setCursor(12,1); lcd.print("S:"); lcd.setCursor(14,1); lcd.print(SR[2]);

}



