/*******************************************
CONTROL UNIT
I2C MASTER
********************************************
HOMS
Hardware-Oriented Microprocessor
Simulator
(C) Panayotis (Panos) Papazoglou
LICENSE: 
Creative Commons
CC BY NC SA
International License
********************************************/
//Please ignore any compilation warnings
//due to function overload (the same function name
//can be used but with different parameters).
//the MCU executes the function which matches
//with parameters

//Define symbolic names for sending commands to other blocks
//The I2C communication is based on one byte transmit/receive
#define READ  65
#define RESET 66
#define SHIFT_L 67
#define SHIFT_R 68
#define DEC 69
#define INC 70  
#define LOAD 71
#define ADD 72

//Set symbolic names for button/LED pins
#define GREEN_button 11
#define RED_button 12
#define GREEN_LED 10
#define RED_LED 9

//Include I2C library
//The I2C function are called using the prefix Wire.
#include <Wire.h>

//Include LCD library
#include <LiquidCrystal.h>

//Set pins for LCD1602 connection
LiquidCrystal lcd(3,4,5,6,7,8);

//Declare variables for the needs of the CU unit (locally)
byte RA=0; 
byte RB=0;
byte RC=0;
byte PC=0;
byte MAR=0;
byte MBR=0;
byte T1=0;
byte T2=0;
byte RS=0;
byte command=0;
byte param=0;
String hexval="";
byte blockADDR;

//Delay between execution process steps
int d=3000;

//Starting function (after reset or power on), runs once
void setup()
{

  //Initialize I2C communication
  Wire.begin();

  //Initialize Serial communication
  //for displaying debugging messages
  Serial.begin(9600);

  //Initialize buttons & LED-buttons
  init_buttons_led();
  //Initialize LCD
  lcd.begin(16, 2);

  //Initial value for entering the while loop
  command=0;

  //Clear registers (content=00)
  //Wait for RED button to start execution
  INIT();

  /****************************
  MAIN EXECUTION PROCESS
  *****************************/
  while (command!=17)
  {
  READ_PC();    //Read PC content for instruction starting address
  PC2MAR();     //Update MAR for accessibg memory address
  FETCH1();     //Fetch 1st byte from memory (instruction code)
  UPDATE_MAR(); //Update MAR for fetching 2nd byte
  FETCH2();     //Fetch 2nd byte (instruction parameter)
  DECODE_EXEC();//Decode instruction and execute
  UPDATE_PC();  //Update PC content for next instruction
  
  //Uncomment the following two code lines if you want a pause
  //between instruction execution

  //lcdtext(0,1,"Press <NEXT>");
  //wait();
  }

  //Out of while loop. The following code is executed when the instruction code is 17
  //which corresponds to STOP instruction
  lcd.clear();
  lcdtext(1,0,"END OF PROGRAM");
  lcdtext(3,1,"EXECUTION");
  
}//End of SETUP


/*******************************************
READ PC (Program Counter), PC block address=4
********************************************/

void READ_PC()
{
  //Display message on LCD
  lcd.clear();
  lcdtext(0,0,"Reading PC REG..");
  delay(d);

  //Read PC content. Store content in PC variable
  PC=READ_REG(4);

  //Display READ results on LCD
  lcd.clear();
  lcd.setCursor(0,0); lcd.print("PC-Prog.Counter");
  lcd.setCursor(0,1); lcd.print("PC10="); 
  lcd.setCursor(5,1); lcd.print(PC);
  lcd.setCursor(9,1); lcd.print("PC16="); 
  hexval =  String(PC, HEX);   
  lcd.setCursor(14,1); lcd.print(hexval);

  //Wait before next step
  delay(d);

} //End of READ_PC

/*******************************************
Store PC contents to register MAR
for accessing memory address
MAR=PC (MAR block address=5)
********************************************/
void PC2MAR()
{
  //Display message on LCD
  lcd.clear();
  lcdtext(0,0,"MAR <-- PC ...");
  delay(d);
 

  //LOAD PC content in MAR register
  WRITE_REG(5,PC);
 
  //Update variable
  MAR=PC;

  //Display message on LCD
  lcd.clear();
  lcdtext(0,0,"Check MAR...");
  delay(d);
}//End of PC2MAR

/*******************************************
FETCH 1st byte from memory, mem[MAR]
Instruction code
********************************************/
void FETCH1()
{
  //Display message on LCD
  lcd.clear();
  lcdtext(0,0,"FETCH byte 1...");
  delay(d);

  //LOAD MAR content in MEM/OUT Unit
  //for accessing the corresponding address
  WRITE_REG(11,MAR);

  //Receive memory content
  //Store content in MBR variable
  blockADDR=11;
  Wire.requestFrom(blockADDR, 1);
  if (Wire.available()) 	      
		MBR = Wire.read();

//Store instruction code in command variable
command=MBR;

//Display information (MAR, MBR) on LCD
lcd.clear();
lcd.setCursor(0,0); lcd.print("MBR<--mem[");
lcd.setCursor(10,0); lcd.print(MAR);
lcd.setCursor(13,0); lcd.print("]");
lcd.setCursor(0,1); lcd.print("(10)="); 
lcd.setCursor(5,1); lcd.print(MBR);
lcd.setCursor(9,1); lcd.print("(16)="); 
hexval =  String(MBR, HEX);   
lcd.setCursor(14,1); lcd.print(hexval);
delay(d);

//Update MBR register from MBR variable
WRITE_REG(6,MBR);

//Display message on LCD
lcd.clear();
lcdtext(0,0,"Check MBR");
delay(d);
lcd.clear();
lcdtext(0,0,"command is"); lcd.setCursor(11,0); lcd.print(command);
delay(d);
} //End of FETCH1


/*******************************************
Updating MAR for accessing 
MAR=PC+1 (MAR address=5)
********************************************/
void UPDATE_MAR()
{
  //Display message on LCD
  lcd.clear();
  lcdtext(0,0,"Updating MAR...");

  //Update MAR register for fetching the next byte from memory
  //(from next address)
  WRITE_REG(5,PC+1);
  delay(d);

  //Update variable
  MAR=PC+1;

  //Display message on LCD
  lcd.clear();
  lcdtext(0,0,"Check MAR");
  delay(d);
}


/*******************************************
FETCH 2nd byte from memory, mem[MAR]
Parameter code
********************************************/
void FETCH2()
{
//Display message on LCD
lcd.clear();
lcdtext(0,0,"FETCH byte 2...");

//LOAD MAR content in MEM/OUT Unit
//for accessing the corresponding address
WRITE_REG(11,MAR);
delay(d);


//Receive memory content
//Store content in MBR variable
blockADDR=11;
Wire.requestFrom(blockADDR, 1);
if (Wire.available()) 	      
		MBR = Wire.read();

//Store parameter code in param variable
param=MBR;

//Display information (MAR, MBR) on LCD
lcd.clear();
lcd.setCursor(0,0); lcd.print("MBR<--mem[");
lcd.setCursor(10,0); lcd.print(MAR);
lcd.setCursor(13,0); lcd.print("]");
lcd.setCursor(0,1); lcd.print("(10)="); 
lcd.setCursor(5,1); lcd.print(MBR);
lcd.setCursor(9,1); lcd.print("(16)="); 
hexval =  String(MBR, HEX);   
lcd.setCursor(14,1); lcd.print(hexval);
delay(d);

//update register MBR
WRITE_REG(6,MBR);

//Display messages on LCD
lcd.clear();
lcdtext(0,0,"Check MBR");
delay(d);
lcd.clear();
lcdtext(0,0,"param is"); lcd.setCursor(11,0); lcd.print(param);
delay(d);
} //END of FETCH2


/*******************************************
Decoding instruction and execution
Instruction block: [command][param]

Implemented Assembly Instructions

CODE  INSTRUCTION
====  ===========
04    LOD A,num
10    INC A
06    MOV B,A
14    ADD A,B
17    STOP

command=instruction code
param=parameter code
********************************************/
//command,param
void DECODE_EXEC()
{

/****************************
LOD A,num (A=num)
*****************************/
if (command==4)
{
  //Display messages on LCD
  lcd.clear();
  lcdtext(0,0,"CMD:LOD A,"); lcd.setCursor(10,0); lcd.print(param);
  lcdtext(0,1,"Executing...");
  delay(d);

  //LOAD param (num) in Register A
  WRITE_REG(1,param);

}

/****************************
INC A (A=A+1)
*****************************/
if (command==10)
{
    //Display messages on LCD
    lcd.clear();
    lcdtext(0,0,"CMD:INC A");
    lcdtext(0,1,"Executing...");
    delay(d);

    //Send INC command to register A 
    blockADDR=1;
    Wire.beginTransmission(blockADDR);
    Wire.write(INC);
    Wire.endTransmission();

    //Display messages on LCD
    lcd.clear();
    lcdtext(0,0,"Check REG-A");
}

/****************************
MOV B,A (B=A)
*****************************/
if (command==6)
{
  //Display messages on LCD
  lcd.clear();
  lcdtext(0,0,"CMD:MOV B,A");
  lcdtext(0,1,"Executing...");
  delay(d);

  //Read from register A
  //Store content to variable RA
  RA=READ_REG(1);
  Serial.print("Read=RA:"); Serial.println(RA);

  //Store register A content (RA) in register B
  WRITE_REG(2,RA);

  //Display message on LCD
  lcd.clear();
  lcdtext(0,0,"Check REG-B");
}

/****************************
ADD A,B
*****************************/
if (command==14)
{
  //Display message on LCD
  lcd.clear();
  lcdtext(0,0,"CMD:ADD A,B");
  delay(d);

//READ from register A, Store in variable RA
RA=READ_REG(1);

//READ from register B, Store in variable RB
RB=READ_REG(2);

//Store register A content in register T1
WRITE_REG(7,RA);

//Store register B content in register T2
WRITE_REG(8,RB);

//Display messages on LCD
lcd.clear();
lcdtext(0,0,"Check T1,T2");
lcdtext(0,1,"ALUEXEC...");
delay(d);

//Send ADD instruction to ALU/SR Unit
byte RES=SEND_ALU(9,ADD);

//Store ALU result in register A
WRITE_REG(1,RES);

//Display message on LCD
lcd.clear();
lcdtext(0,0,"Check RA,ALU,SR");
delay(d);
}


}//End of DECODE_EXEC

/*******************************************
Update PC register content
for fetching next instruction bytes
from memory
********************************************/
void UPDATE_PC()
{

    //Send two INC commands to PC register
    Wire.beginTransmission(4);
    Wire.write(INC);
    Wire.endTransmission();
    delay(50);
    Wire.beginTransmission(4);
    Wire.write(INC);
    Wire.endTransmission();
    delay(50);

    //Update variable
    PC=PC+2;
}

void loop()
{
  //Nothing here
}



/*******************************************
Initialize PULL-UP resistors
Every button is activated when
a PULL-UP pin goes to LOW (GND)

button-state=1 inactive
button-state=0 active

Activate LEDs inside buttons
********************************************/
void init_buttons_led()
{
  pinMode(GREEN_button,INPUT_PULLUP);
  pinMode(RED_button,INPUT_PULLUP);

  pinMode(GREEN_LED, OUTPUT);
  pinMode(RED_LED, OUTPUT);

  digitalWrite(GREEN_LED,HIGH);
  digitalWrite(RED_LED,HIGH);


}

/*******************************************
Pause
wait on GREEN button press
********************************************/
void wait()
{
  while(digitalRead(GREEN_button)==HIGH) {;}
  delay(250);
}

/*******************************************
Pause
wait on RED button press
********************************************/
void wait_red()
{
  while(digitalRead(RED_button)==HIGH) {;}
  delay(250);
}

/*******************************************
Function for displaying Strings on LCD
********************************************/
void lcdtext(int col, int lin, String text)
{
  lcd.setCursor(col,lin);
  lcd.print(text);
  
} 

/*******************************************
Clear all registers (set content=0)

Register address = 1 = Register A
Register address = 2 = Register B
Register address = 3 = Register C
Register address = 4 = Register PC
Register address = 5 = Register MAR
Register address = 6 = Register MBR
Register address = 7 = Register T1
Register address = 8 = Register T2
********************************************/
void clear_REGS()
{
  for(int i=1;i<=8;i++)
  {
    Wire.beginTransmission(i);
    Wire.write(RESET);
    Wire.endTransmission();
    delay(150);
  }
}

/*******************************************
Display messages on LCD
and call clear register function
********************************************/
void INIT()
{
lcd.clear();
lcdtext(0,0,"RESET REGs..");
lcdtext(0,1,"Please Wait...");
clear_REGS();
delay(d);


lcd.clear();
lcdtext(0,0,"Press RED button");
lcdtext(0,1,"to START");
wait_red();

}



/*******************************************
Read register content
Send READ command
Register address = addr
Return content
********************************************/
byte READ_REG(byte addr)
{
byte ans;

  Wire.beginTransmission(addr);
  Wire.write(READ);
  Wire.endTransmission();
  delay(200);
  //Receive answer
  Wire.requestFrom(addr, 1);
  if (Wire.available()) 	      
		ans = Wire.read();

return ans;
}

/*******************************************
Send command to ALU/SR unit
Address = addr
command to ALU/SR = instruction
Return ALU result
********************************************/
byte SEND_ALU(byte addr, byte instruction)
{
byte ans;

  Wire.beginTransmission(addr);
  Wire.write(instruction);
  Wire.endTransmission();
  delay(300);
  //Receive answer (PC content)
  
  Wire.requestFrom(addr, 1);
  if (Wire.available()) 	      
		ans = Wire.read();

return ans;
}


/*******************************************
Load a number to a specific register
Register address = addr
Number = data
********************************************/
void WRITE_REG(byte addr, byte data)
{

Wire.beginTransmission(addr);
Wire.write(LOAD);
Wire.endTransmission();
delay(200);
Wire.beginTransmission(addr);
Wire.write(data);
Wire.endTransmission();
}
