/*******************************************
MEMORY & OUTPUT UNIT
********************************************
HOMS
Hardware-Oriented Microprocessor
Simulator
(C) Panayotis (Panos) Papazoglou
LICENSE: 
Creative Commons
CC BY NC SA
International License
********************************************/

/*******************************************
LIBRARIES for the TFT 2.4 Inch screen
********************************************/
#include <Adafruit_GFX.h>
#include <MCUFRIEND_kbv.h>

//Include I2C library
#include <Wire.h>
//Object declaration for accessing TFT library functions
//Every function call starts with the prefix tft.
MCUFRIEND_kbv tft;

//TFT Color codes for easy access
#define BLACK   0x0000
#define BLUE    0x001F
#define RED     0xF800
#define GREEN   0x07E0
#define CYAN    0x07FF
#define MAGENTA 0xF81F
#define YELLOW  0xFFE0
#define WHITE   0xFFFF

//A byte is read from memory when the LOAD command is received
#define LOAD 71

//8bit switch pins
int bitsw[]={28,26,24,22,38,36,34,32}; 

//Button pins
int button[]={42,44,46}; 

//Button LEDs pins
int buttonLED[]={48,50,52}; 

//Button state
int bstate[]={0,0,0}; 

//Switch state 1=inactive (due to PULL_UP resistors)
int swstate[]={1,1,1,1,1,1,1,1}; 

//Memory: 256 locations, starting address 0, ending address 255 (FF hex)
int mem[256]; 

//Current selected address
int addr=0;  

//Current data for storing in memory
int data=0; 

//Keep previous values for updating screen only if values are changed
int addrprev=0, dataprev=0; 

//Number in dec format
float dec=0; 

//Integer dec number
int idec=0; 

//Previous dec value (round is used before decprev update)
int decprev=0; 

//Previous hex value (round is used before hexprev update)
String fhexprev; 

//Received data from control unit
byte rV=0; 

//Previous received data from control unit
byte rVprev=0;

//Flag that is activated when an address is received
byte ADDRflag=0;

//Declare next array if you want to have a preloaded code in memory
byte prog[]={4, 8, 10, 0, 6, 0, 14, 0, 17, 0};

//Starting function (after reset or power on), runs once
void setup(void)
{
  
   //initialize I2C communication (address 11)
  Wire.begin(11);

  //Enable ISR routines for receiving bytes and requests
  Wire.onReceive(onReceive);
  Wire.onRequest(onRequest);

  //Initialize Serial communication
  //for displaying debugging messages
  Serial.begin(9600);

  //initialize switches and buttons
  init_sw_buttons();

  //initialize button LEDs
  init_buttonLEDs(); 

  //initialize memory
  init_mem(256);  

  //Store prog[] bytes in memory (preload code)
  set_prog();

  //initialize TFT
  init_TFT();   

  //display welcome messages on TFT
  TFT_welcome();    
}


/*******************************************
Main function (always active)
********************************************/
void loop(void)
{ 
/*******************************************
Read switch states (8 switches)
store each state in swstate[] array
use bitsw[] array for switch pins
********************************************/
  for(int i=0;i<8;i++)
    swstate[i]=digitalRead(bitsw[i]);

  //print dec and hex value on screen (based on switches state)
  print_dechex();

/*******************************************
Read button states (3 buttons)
store each state in bstate[] array
use button[] array for button pins
********************************************/
for(int i=0;i<3;i++)
    bstate[i]=digitalRead(button[i]);

/***********************************************************************************
Check button states from bstate[] array
pins:
button[0]=GREEN, button[1]=RED RIGHT, button[2]=RED LEFT
if the RED LEFT button is pressed, the dec value is stored as address
if the RED RIGHT button is pressed, the dec value is tored as data (content)
if the GREEN button is pressed, the user wants to store data in the address addr
or wants to display memory contents on the screen
************************************************************************************/
if (bstate[2]==LOW) addr=round(dec); if (bstate[1]==LOW) data=round(dec);
if (bstate[0]==LOW) 
{
  //If both addr and data values are 128, the memory contents are displayed on the TFT screen
  if ((addr==128) && (data==128))
  {
    //Clear TFT screen
    clear_TFT(BLACK);

    //Display titles for the two DATA columns
    print_text(YELLOW, RED,2,10,10,"ADDR"); print_text(YELLOW, RED,2,60,10,"DATA");
    print_text(YELLOW, RED,2,110,10,"ADDR"); print_text(YELLOW, RED,2,160,10,"DATA");

    //Variables for moving to the desired X,Y position
    int x=10, y=25; 

    //Loop for display memory contents, 1st column from 0 to 10, 2nd column from 11 to 21
    for(int i=0;i<11;i++)
    {
      //Display address for first range
      print_text(WHITE, BLACK,2,x,y,String(i));

      //Display address content for first range
      print_text(WHITE, BLACK,2,x+50,y,String(mem[i]));

      //Display address for second range
      print_text(WHITE, BLACK,2,x+100,y,String(i+11));

      //Display address content for second range
      print_text(WHITE, BLACK,2,x+150,y,String(mem[i+11]));
      
      //Update Y position for changing line
      y+=20; 

    }

    //Display EXIT label
    print_text(WHITE, BLUE,2,220,180,"EXIT");

    //Wait on that screen until green button is pressed
    while(digitalRead(button[0])==HIGH) {;}

    //Clear TFT screen
    clear_TFT(BLACK);

  }
  else
  //the user has selected to store data value in memory
  {

  //Display SET for confirming green button press
  print_text(YELLOW, RED,3,220,180,"SET");

  //Store data value in memory (mem[] array)
  mem[addr]=data;

  delay(100);

  //Display OK. The store process is successful
  print_text(WHITE, BLUE,3,220,180,"OK!");
  delay(500);
  }
}

//If a new ADDRESS is entered, clear for the new value 
if (addr!=addrprev) tft.fillRect(50, 180,100,35,BLACK);

//If new DATA are entered, clear for the new value 
if (data!=dataprev) tft.fillRect(150, 180,100,35,BLACK);

//Store current values of data and address before the next update
dataprev=data; addrprev=addr;

//Display current address value (updated with the RED LEFT button based on switches state)
print_text(WHITE, RED,3,10,180,"A:");
print_text(YELLOW, RED,3,50,180,String(addr));

//Display current data value (updated with the RED RIGHT button based on switches state)
print_text(WHITE, RED,3,120,180,"D:");
print_text(YELLOW, RED,3,150,180,String(data));

//Display label for GREEN button
print_text(BLACK, GREEN,3,220,180,"SET");

//Delay before return to loop starting point
delay(30);  
}

/*******************************************
Initialize switches and buttons
using microcontroller PULL_UP resistors.
If a switch or a button is inactive
the corresponding state is HIGH
********************************************/
void init_sw_buttons()
{
  for(int i=0;i<8;i++)
    pinMode(bitsw[i],INPUT_PULLUP);

  for(int i=0;i<3;i++)
    pinMode(button[i],INPUT_PULLUP);
}


/*******************************************
Initialize button LEDs.
Set all pins to HIGH.
Lighting the LEDs
********************************************/
void init_buttonLEDs()
{
  for(int i=0;i<3;i++)
  {
    pinMode(buttonLED[i],OUTPUT);
    digitalWrite(buttonLED[i],HIGH);
  }

}

/*******************************************
Initialize TFT screen
rotate coordinate system to use
screen as landscape
********************************************/
void init_TFT()
{
   uint16_t ID = tft.readID();
    tft.begin(ID);
    tft.setRotation(1);            //PORTRAIT
    clear_TFT(BLACK);
}

/*******************************************
Display welcome messages
********************************************/
void TFT_welcome()
{
  clear_TFT(BLUE);
  int x=30;
  print_text(WHITE,BLUE, 3, x,30,"Hardware");
  print_text(WHITE,BLUE, 3, x,60,"based");
  print_text(WHITE,BLUE, 3, x,90,"Microprocessor");
  print_text(WHITE,BLUE, 3, x,120,"Simulator V1.0");
  print_text(YELLOW,BLUE, 2, 20,160,"(C) P. PAPAZOGLOU, 2023");
  print_text(WHITE, BLUE,2,60,200,"-- PRESS GREEN --");

  //Wait on that screen until green button is pressed
  while(digitalRead(button[0])==HIGH) {;}
 
  clear_TFT(BLACK);
 
    }

/*******************************************
Display DEC and HEX value on screen
based on 8 states of the switches
8 states = 8 bits
********************************************/
void print_dechex()
{ 
  //Initialize dec value
  dec=0;
  
  //Exponent of two starting from the MSB bit
  int p=7;  
  
  //Variable for bit value based on switches state (switch=0=active, bit=1)
  int b; 
  
  //Variables for moving to the desired X,Y locations
  int x=10, y=50;

  //Display label for the binary representation
  print_text(WHITE, BLACK, 5, 10, 10, "BIN");

  //Display 8bit value and calculate the decimal value
  for(int i=0;i<8;i++)
  {
    //Set bit value based on switch state
    if (swstate[i]==LOW) b=1; else b=0;

    //Display current bit
    print_text(YELLOW, BLACK, 4, x, y, String(b));
    
    //Update X location for displaying the next right bit
    x+=30;

    //Calculate dec value for current bit (b*2^p)
    float value=b*pow(2,p);

    //Update sum for the whole number
    dec=dec+value; 

    //Decrease the exponent (based on bit position in the number)
    p--;
  }

  //idec is the integer representation of dec (round function is used 
  //for value integrity due to type conversion)
  idec=round(dec);

  //Convert decimal value to hexadecimal (hex)
  String fhex =  String(idec, HEX);

  //For a new hex value, clear display field
  if (fhex.compareTo(fhexprev)!=0) tft.fillRect(180, 120,50,40,BLACK);

  //Keep current hex value (as previous)
  fhexprev=fhex;

  //Display label and hex value on screen
  print_text(WHITE, BLACK, 4, 150, 90, "HEX");
  print_text(YELLOW, BLACK, 4, 170, 130, fhex);

  //For a new dec value, clear display field
  if (idec!=decprev) tft.fillRect(10, 120,215,40,BLACK);

  //Keep current dec value (as previous)
  decprev=idec;

  //Display label and dec value on screen
  print_text(WHITE, BLACK, 5, 10, 90, "DEC");
  print_text(YELLOW, BLACK, 4, 10, 130, String(idec));
}


/*******************************************
display text (String) on the TFT screen

fcolor: text color
bcolor: text background color
text_size: size of text
xpos: X position on screen
ypos: Y position on screen
text: text to be displayed
********************************************/
void print_text(int fcolor, int bcolor, int text_size, int xpos, int ypos, String text)
{
  tft.setTextColor(fcolor, bcolor);
    tft.setTextSize(text_size);
    tft.setCursor(xpos, ypos);
    tft.print(text);
}


/*******************************************
clear TFT screen, fill with color
********************************************/
void clear_TFT(int color)
{
tft.fillScreen(color);

}

/*******************************************
initialize memory with size of locations

locations: total array locations
********************************************/
void init_mem(int locations)
{
  for(int i=0;i<locations;i++)
    mem[i]=0;
}

/*******************************************
When data are received from control unit
********************************************/
void onReceive(int a) 

{ 
  rV=Wire.read();
  Serial.print("rV:"); Serial.println(rV);

  //If previous command is LOAD, then the current byte represents an address
  if (rVprev==LOAD) ADDRflag=1;
  rVprev=rV;
}

/*******************************************
When data are requested from control unit
********************************************/
void onRequest()
{
    //If the contents of an address are requested
    if (ADDRflag==1) 
    {
      Serial.println("ADDRflag=1");
      Serial.print("rV:");Serial.print(rV);
      Serial.print(", content:"); Serial.println(mem[rV]);
      Wire.write(mem[rV]); 
      ADDRflag=0;
      rVprev=0;
    }
}

/*******************************************
Preload program bytes in memory.
Update mem[] array from prog[] array
********************************************/
void set_prog()
{
  for(int i=0;i<10;i++) mem[i]=prog[i];
}