/*******************************************
REGISTER UNIT
All registers execute the same code,
the only difference is the I2C address
********************************************
HOMS
Hardware-Oriented Microprocessor
Simulator
(C) Panayotis (Panos) Papazoglou
LICENSE: 
Creative Commons
CC BY NC SA
International License
********************************************/
//Change address based on block that is used
#define ADDRESS 2

//Include I2C library
#include <Wire.h>


/*******************************************
Declare PINS for seven segment display module
********************************************/
#define Data 2
#define Clock 3
#define Load 4

//Define symbolic names for instructions codes
//between blocks
#define READ  65
#define RESET 66
#define SHIFT_L 67
#define SHIFT_R 68
#define DEC 69
#define INC 70  
#define LOAD 71

/*******************************************
Segment status ON/OFF for one digit [0, F]
********************************************/
byte hex[16]=
{
   0b00000011,  //0
   0b10011111,  //1
   0b00100101,  //2
   0b00001101,  //3
   0b10011001,  //4
   0b01001001,  //5
   0b01000001,  //6
   0b00011111,  //7
   0b00000001,  //8
   0b00001001,  //9
   0b00010001,  //a
   0b11000001,  //b
   0b11100101,  //c
   0b10000101,  //d
   0b01100001,  //e
   0b01110001   //f
};

//byte that is received via I2C communication
byte rV; 

//Previous received byte. Is used when a sequence of bytes
//has to be checked
byte rVprev=0;

//Initial register content
byte REG=0x00;

//Starting function (after reset or power on), runs once
void setup() {

  //Initialize Serial communication
  //for displaying debugging messages
  Serial.begin(9600);
  
  //Initialize seven segment module
  init_display();
  
  //Display register content
  display(REG);

  //Initialize I2C communication:
  //Set address
  //Enable ISR for receiving bytes and requests
  Wire.begin(ADDRESS);
  Wire.onReceive(onReceive);
  Wire.onRequest(onRequest);
}

void loop() 
{
  //Nothing here

}

/*******************************************
Action Routine that is activated when 
a byte is received 
********************************************/
void onReceive(int a) 

{ 
  
  //Read received byte
  rV=Wire.read();

  //If previous byte is 71 (LOAD), then the current byte
  //is the value that will be loaded in register
  if (rVprev==LOAD) REG=rV;

  //Update variable for previous value
  rVprev=rV;
 
  //Actions based on received byte
  if (rV==INC) REG++;
  if (rV==DEC) REG--;
  if (rV==SHIFT_R) REG=REG>>1;
  if (rV==SHIFT_L) REG=REG<<1;
  if (rV==RESET) REG=0;

  //Display register content after performed action
   display(REG);

}

/*******************************************
Action Routine that is activated when 
a request is received 
********************************************/
void onRequest()
{

  //If the received byte represents 
  //the READ command, then the 
  //register content is sent as answer
  if (rV==READ) Wire.write(REG);

  //Display register content
  display(REG);
}

/*******************************************
Display register content as HEX number on
seven segment module 
Digit manipulation for left and right
seven segment display unit 
********************************************/
void display(byte REGnum)
{

//Convert decimal number REGnum to HEX
String Shex=String(REGnum,HEX);

//Variables for left and right digit
char LL;
char RR;

//If the hex number has only one digit, then 
//the left digit is zero ('0') and the right digit
//is the first character of the string Shex
if (Shex.length()==1)
  { LL='0'; RR=Shex[0];}

//Otherwise, update left and right digit variables from
//the whole string Shex
else
{
  LL=Shex[0];
  RR=Shex[1];
}

//variables for accessing later the hex[] array 
//in order to activate the correct segment
//on seven segment module
byte left,right;

//Serial.print("LL="); Serial.println(LL);
//Serial.print("RR="); Serial.println(RR);

//Translate one digit numbers from string form
//to single numbers for accessing hex[] array
//The calculations are based on the corresponding ASCII
//code for each string digit
//for example, character 'a' has the ASCII code 97
//the calculation is 'a'-87=97-87=10. The number 10
//will be used as index in the hex[] array (hex[10])
//for displaying 'A'
if (LL>='a' && LL<='f') left=LL-87; else
    if (LL>='0' && LL<='9') left=LL-48;

  if (RR>='a' && RR<='f') right=RR-87; else
      if (RR>='0' && RR<='9') right=RR-48;


//Serial.print(left); Serial.print(", "); Serial.println(right);

  //Display final hex number to seven segment display module 
  digitalWrite(Load,LOW);
  shiftOut(Data,Clock, LSBFIRST, hex[right]);
  shiftOut(Data,Clock, LSBFIRST, hex[left]);
  digitalWrite(Load,HIGH);

}

/*******************************************
Initialize
seven segment module 
********************************************/
void init_display()
{
  pinMode(Data,OUTPUT);
  pinMode(Clock,OUTPUT);
  pinMode(Load, OUTPUT);
}